[formatters]
keys=simple

[loggers]
keys=root,nodepool,openstack,kazoo,keystoneauth,novaclient

[handlers]
keys=file

[logger_root]
level=WARNING
handlers=file

[logger_nodepool]
level=DEBUG
handlers=file
qualname=nodepool
propagate=0

[logger_openstack]
level=DEBUG
handlers=file
qualname=openstack
propagate=0

[logger_keystoneauth]
level=DEBUG
handlers=file
qualname=keystoneauth
propagate=0

[logger_novaclient]
level=DEBUG
handlers=file
qualname=novaclient
propagate=0

[logger_kazoo]
level=INFO
handlers=file
qualname=kazoo
propagate=0

[handler_file]
level=DEBUG
class=FileHandler
formatter=simple
args=('{{ log_file }}', 'w')

[formatter_simple]
format=%(asctime)s %(levelname)s %(name)s: %(message)s
datefmt=
