version: '3.0'
services:
  nodepool-builder:
    image: zuul/nodepool-builder{{ nodepool_container_tag|default('') }}
    network_mode: host
    restart: always

    privileged: true

    environment:
      - DEBUG=1

    volumes:
      # cloud config
      - /etc/openstack:/etc/openstack:ro
      # nodepool config
      - /etc/nodepool:/etc/nodepool:ro
      # large dib storage area
      - /opt/dib:/opt/dib
      # for dib to get source for glean installs, etc.
      - /home/zuul:/home/zuul
      # logs (builder + dib build logs under /build)
      - /var/log/nodepool:/var/log/nodepool
      # devstack tls-proxy puts CA here that is referenced by cloud config
      - /opt/stack/data:/opt/stack/data:ro

  nodepool-launcher:
    image: zuul/nodepool-launcher{{ nodepool_container_tag|default('') }}
    network_mode: host
    restart: always

    environment:
      - DEBUG=1

    volumes:
      # cloud config
      - /etc/openstack:/etc/openstack:ro
      # nodepool config
      - /etc/nodepool:/etc/nodepool:ro
      # logs (builder + dib build logs under /build)
      - /var/log/nodepool:/var/log/nodepool
      # devstack tls-proxy puts CA here that is referenced by cloud config
      - /opt/stack/data:/opt/stack/data:ro
